import json
from datetime import datetime, timedelta
from uuid import uuid4

from apscheduler.executors.pool import ProcessPoolExecutor
from apscheduler.schedulers.blocking import BlockingScheduler as APScheduler
from botocore.exceptions import ClientError
from probator import app_config, get_local_aws_session, AWS_REGIONS
from probator.config import dbconfig
from probator.constants import SchedulerStatus, ConfigOption
from probator.database import db
from probator.exceptions import ProbatorError, SchedulerError, ProbatorFatalError
from probator.plugins import CollectorType, BaseScheduler
from probator.plugins.types.accounts import BaseAccount, AWSAccount
from probator.schema.base import SchedulerBatch, SchedulerJob
from probator.utils import get_hash
from probator.wrappers import retry


class SQSScheduler(BaseScheduler):
    name = 'SQS Scheduler'
    ns = 'scheduler_sqs'
    options = (
        ConfigOption('queue_region', 'us-west-2', 'string', 'Region of the SQS Queues'),
        ConfigOption('job_queue_url', '', 'string', 'URL of the SQS Queue for pending jobs'),
        ConfigOption('status_queue_url', '', 'string', 'URL of the SQS Queue for worker reports'),
        ConfigOption(
            'job_delay',
            2,
            'float',
            'Time between each scheduled job, in seconds. Can be used to avoid spiky load during execution of tasks'
        ),
    )

    def __init__(self):
        """Initialize the SQSScheduler, setting up the process pools, scheduler and connecting to the required SQS Queues"""
        super().__init__()

        self.pool = ProcessPoolExecutor(1)
        self.scheduler = APScheduler(
            threadpool=self.pool,
            job_defaults={
                'coalesce': True,
                'misfire_grace_time': 30
            }
        )

        session = get_local_aws_session()
        sqs = session.resource('sqs', self.dbconfig.get('queue_region', self.ns))

        job_queue_url = self.dbconfig.get('job_queue_url', self.ns)
        status_queue_url = self.dbconfig.get('status_queue_url', self.ns)

        if job_queue_url.endswith('.fifo') and status_queue_url.endswith('.fifo'):
            self.is_fifo = True

        elif not job_queue_url.endswith('.fifo') and not status_queue_url.endswith('.fifo'):
            self.is_fifo = False

        else:
            raise ProbatorFatalError('Cannot mix fifo and non-fifo queues')

        self.job_queue = sqs.Queue(job_queue_url)
        self.status_queue = sqs.Queue(status_queue_url)

    def execute_scheduler(self):
        """Execute the scheduler

        Main entry point for the scheduler. This method will start two scheduled jobs, `schedule_jobs` which takes
        care of scheduling the actual SQS messaging and `process_status_queue` which will track the current status
        of the jobs as workers are executing them

        Returns:
            `None`
        """
        try:
            self.scheduler.add_job(
                self.cleanup,
                trigger='cron',
                name='cleanup',
                hour=3,
                minute=0,
                second=0
            )

            # Schedule periodic scheduling of jobs
            self.scheduler.add_job(
                self.schedule_jobs,
                trigger='interval',
                name='schedule_jobs',
                minutes=15,
                start_date=datetime.now() + timedelta(seconds=1)
            )

            self.scheduler.add_job(
                self.process_status_queue,
                trigger='interval',
                name='process_status_queue',
                seconds=30,
                start_date=datetime.now() + timedelta(seconds=5),
                max_instances=1
            )

            self.scheduler.start()

        except KeyboardInterrupt:
            self.scheduler.shutdown()

    def list_current_jobs(self):
        """Return a list of the currently scheduled jobs in APScheduler

        Returns:
            `dict` of `str`: :obj:`apscheduler/job:Job`
        """
        jobs = {}
        for job in self.scheduler.get_jobs():
            if job.name not in ('schedule_jobs', 'process_status_queue'):
                jobs[job.name] = job

        return jobs

    def schedule_jobs(self):
        """Schedule or remove jobs as needed.

        Checks to see if there are any jobs that needs to be scheduled, after refreshing the database configuration
        as well as the list of collectors and auditors.

        Returns:
            `None`
        """
        self.dbconfig.reload_data()
        self.collectors = {}
        self.auditors = []
        self.load_plugins()

        _, accounts = BaseAccount.search(include_disabled=False)
        current_jobs = self.list_current_jobs()
        new_jobs = []
        batch_id = str(uuid4())

        batch = SchedulerBatch()
        batch.batch_id = batch_id
        batch.status = SchedulerStatus.PENDING
        db.session.add(batch)
        db.session.commit()

        start = datetime.now() + timedelta(seconds=1)
        job_delay = dbconfig.get('job_delay', self.ns, 0.5)

        # region Global collectors (non-aws)
        if CollectorType.GLOBAL in self.collectors:
            for worker in self.collectors[CollectorType.GLOBAL]:
                job_name = get_hash(worker)

                if job_name in current_jobs:
                    continue

                self.log.info('Scheduling global {} worker every {} minutes'.format(
                    worker.name,
                    worker.interval
                ))

                self.scheduler.add_job(
                    self.send_worker_queue_message,
                    trigger='interval',
                    name=job_name,
                    minutes=worker.interval,
                    start_date=start,
                    kwargs={
                        'batch_id': batch_id,
                        'job_name': job_name,
                        'entry_point': worker.entry_point,
                        'worker_args': {}
                    }
                )
                start += timedelta(seconds=job_delay)
        # endregion

        # region AWS collectors
        aws_accounts = list(filter(lambda x: x.account_type == AWSAccount.account_type, accounts))
        if CollectorType.AWS_ACCOUNT in self.collectors:
            for worker in self.collectors[CollectorType.AWS_ACCOUNT]:
                for account in aws_accounts:
                    job_name = get_hash((account.account_name, worker))
                    if job_name in current_jobs:
                        continue

                    new_jobs.append(job_name)
                    self.log.info('Scheduling {} worker every {} minutes for {}'.format(
                        worker.name,
                        worker.interval,
                        account.account_name
                    ))

                    self.scheduler.add_job(
                        self.send_worker_queue_message,
                        trigger='interval',
                        name=job_name,
                        minutes=worker.interval,
                        start_date=start,
                        kwargs={
                            'batch_id': batch_id,
                            'job_name': job_name,
                            'entry_point': worker.entry_point,
                            'worker_args': {
                                'account': account.account_name
                            }
                        }
                    )
                    start += timedelta(seconds=job_delay)

        if CollectorType.AWS_REGION in self.collectors:
            for worker in self.collectors[CollectorType.AWS_REGION]:
                for region in AWS_REGIONS:
                    for account in aws_accounts:
                        job_name = get_hash((account.account_name, region, worker))

                        if job_name in current_jobs:
                            continue

                        new_jobs.append(job_name)
                        self.log.info('Scheduling {} worker every {} minutes for {}/{}'.format(
                            worker.name,
                            worker.interval,
                            account.account_name,
                            region
                        ))

                        self.scheduler.add_job(
                            self.send_worker_queue_message,
                            trigger='interval',
                            name=job_name,
                            minutes=worker.interval,
                            start_date=start,
                            kwargs={
                                'batch_id': batch_id,
                                'job_name': job_name,
                                'entry_point': worker.entry_point,
                                'worker_args': {
                                    'account': account.account_name,
                                    'region': region
                                }
                            }
                        )
                        start += timedelta(seconds=job_delay)
        # endregion

        # region Auditors
        if app_config.log_level == 'DEBUG':
            audit_start = start + timedelta(seconds=5)
        else:
            audit_start = start + timedelta(minutes=5)

        for worker in self.auditors:
            job_name = get_hash((worker,))
            if job_name in current_jobs:
                continue

            new_jobs.append(job_name)

            self.log.debug('Scheduling {} auditor every {} minutes to start at {}'.format(
                worker.name,
                worker.interval,
                audit_start
            ))

            self.scheduler.add_job(
                self.send_worker_queue_message,
                trigger='interval',
                name=job_name,
                minutes=worker.interval,
                start_date=audit_start,
                kwargs={
                    'batch_id': batch_id,
                    'job_name': job_name,
                    'entry_point': worker.entry_point,
                    'worker_args': {}
                }
            )
            audit_start += timedelta(seconds=job_delay)
        # endregion

    def send_worker_queue_message(self, *, batch_id, job_name, entry_point, worker_args, retry_count=0):
        """Send a message to the `worker_queue` for a worker to execute the requests job

        Args:
            batch_id (`str`): Unique ID of the batch the job belongs to
            job_name (`str`): Non-unique ID of the job. This is used to ensure that the same job is only scheduled
            a single time per batch
            entry_point (`dict`): A dictionary providing the entry point information for the worker to load the class
            worker_args (`dict`): A dictionary with the arguments required by the worker class (if any, can be an
            empty dictionary)
            retry_count (`int`): The number of times this one job has been attempted to be executed. If a job fails to
            execute after 3 retries it will be marked as failed

        Returns:
            `None`
        """
        try:
            job_id = str(uuid4())
            args = {
                'MessageBody': json.dumps({
                    'batch_id': batch_id,
                    'job_id': job_id,
                    'job_name': job_name,
                    'entry_point': entry_point,
                    'worker_args': worker_args,
                }),
                'MessageAttributes': {
                    'RetryCount': {
                        'StringValue': str(retry_count),
                        'DataType': 'Number'

                    }
                }
            }

            if self.is_fifo:
                args.update({
                    'MessageDeduplicationId': job_id,
                    'MessageGroupId': batch_id,
                })

            self.job_queue.send_message(**args)

            if retry_count == 0:
                job = SchedulerJob()
                job.job_id = job_id
                job.batch_id = batch_id
                job.status = SchedulerStatus.PENDING
                job.data = worker_args

                db.session.add(job)
                db.session.commit()

        except Exception as ex:
            self.log.exception(f'Error when processing worker task: {ex}')

    def execute_worker(self):
        """Retrieve a message from the `worker_queue` and process the request.

        This function will read a single message from the `worker_queue` and load the specified `EntryPoint`
        and execute the worker with the provided arguments. Upon completion (failure or otherwise) a message is sent
        to the `status_queue` information the scheduler about the return status (success/failure) of the worker

        Returns:
            `None`
        """
        try:
            try:
                messages = self.job_queue.receive_messages(
                    MaxNumberOfMessages=1 if self.is_fifo else 10,
                    MessageAttributeNames=('RetryCount',)
                )

            except ClientError:
                self.log.exception('Failed fetching messages from SQS queue')
                return

            if not messages:
                self.log.debug('No pending jobs')
                return

            for message in messages:
                retry_count = int(message.message_attributes['RetryCount']['StringValue'])

                data = json.loads(message.body)
                try:
                    # SQS FIFO queues will not allow another thread to get any new messages until the messages
                    # in-flight are returned to the queue or deleted, so we remove the message from the queue as
                    # soon as we've loaded the data
                    self.send_status_message(data['job_id'], SchedulerStatus.STARTED)
                    message.delete()

                    cls = self.get_class_from_ep(data['entry_point'])
                    worker = cls(**data['worker_args'])
                    worker.run()

                    self.send_status_message(data['job_id'], SchedulerStatus.COMPLETED)
                except ProbatorError as ex:
                    if isinstance(ex, ProbatorFatalError):
                        # Fail immediately if its a Fatal error
                        self.send_status_message(data['job_id'], SchedulerStatus.FAILED)

                    elif retry_count >= 3:
                        self.send_status_message(data['job_id'], SchedulerStatus.FAILED)
                    else:
                        self.send_worker_queue_message(
                            batch_id=data['batch_id'],
                            job_name=data['job_name'],
                            entry_point=data['entry_point'],
                            worker_args=data['worker_args'],
                            retry_count=retry_count + 1
                        )

        except KeyboardInterrupt:
            self.log.info('Shutting down worker thread')

    @retry
    def send_status_message(self, object_id, status):
        """Send a message to the `status_queue` to update a job's status.

        Returns `True` if the message was sent, else `False`

        Args:
            object_id (`str`): ID of the job that was executed
            status (:obj:`SchedulerStatus`): Status of the job

        Returns:
            `bool`
        """
        try:
            body = json.dumps({
                'id': object_id,
                'status': status
            })

            args = {
                'MessageBody': body
            }

            if self.is_fifo:
                args.update({
                    'MessageGroupId': 'job_status',
                    'MessageDeduplicationId': get_hash((object_id, status))
                })

            self.status_queue.send_message(**args)
            return True
        except Exception as ex:
            print(ex)
            return False

    @retry
    def process_status_queue(self):
        """Process all messages in the `status_queue` and check for any batches that needs to change status

        Returns:
            `None`
        """
        self.log.debug('Start processing status queue')
        while True:
            messages = self.status_queue.receive_messages(MaxNumberOfMessages=10)

            if not messages:
                break

            for message in messages:
                data = json.loads(message.body)
                job = SchedulerJob.get(data['id'])
                try:
                    if job and job.update_status(data['status']):
                        db.session.commit()
                except SchedulerError as ex:
                    if hasattr(ex, 'message') and ex.message == 'Attempting to update already completed job':
                        pass

                message.delete()

        # Close any batch that is now complete
        open_batches = db.SchedulerBatch.find(SchedulerBatch.status < SchedulerStatus.COMPLETED)
        for batch in open_batches:
            open_jobs = list(filter(lambda x: x.status < SchedulerStatus.COMPLETED, batch.jobs))
            if not open_jobs:
                open_batches.remove(batch)
                batch.update_status(SchedulerStatus.COMPLETED)
                self.log.debug('Closed completed batch {}'.format(batch.batch_id))
            else:
                started_jobs = list(filter(lambda x: x.status > SchedulerStatus.PENDING, open_jobs))
                if batch.status == SchedulerStatus.PENDING and len(started_jobs) > 0:
                    batch.update_status(SchedulerStatus.STARTED)
                    self.log.debug('Started batch manually {}'.format(batch.batch_id))

        # Check for stale batches / jobs
        for batch in open_batches:
            if batch.started < datetime.now() - timedelta(hours=2):
                self.log.warning('Closing a stale scheduler batch: {}'.format(batch.batch_id))
                for job in batch.jobs:
                    if job.status < SchedulerStatus.COMPLETED:
                        job.update_status(SchedulerStatus.ABORTED)
                batch.update_status(SchedulerStatus.ABORTED)
        db.session.commit()
